"""
 #  Mod_Dal(Distributions) by pyTorch
 @  E.C.Ares @OpenGI; OpenAI
 !  MIT DIVIƷON
 `  Modify simple Distributions for compatibility
"""


from  argparse import Namespace as Nym
import      torch               as nmb
import      torch.nn            as nmm
import      torch.nn.functional as nmf
import      torch.distributions as dal
from        torch.distributions.utils import _sum_rightmost

#msd=nmb.std
log=nmb.log
exp=nmb.exp
mex=nmb.max
mev=nmb.min
abs=nmb.abs
sum=nmb.sum
met=nmb.mean
med=nmb.mode
mad=nmb.clip
sig=nmb.sign
rsq=nmb.sqrt
rod=nmb.round
mix=nmb.argmax
miv=nmb.argmin
mar=nmb.softmax
lan=nmb.linspace
logsumexp=nmb.logsumexp
symlog = lambda x:sig(x)* log(abs(x) +1.0)
symexp = lambda x:sig(x)*(exp(abs(x))-1.0)



# --- Dals --- #

# 重参数化技巧: 将随机变量的采样转化为确定性函数加上随机噪声的方法，就可微了
#dal.independent.Independent(
#dal.uniform.Uniform


# ---
class SampleDist:
    def __init__(ego, dist, samples=100):
        ego._dist = dist
        ego._samples = samples

    @property
    def name(ego):
        return "SampleDist"

    def __getattr__(ego, name):return getattr(ego._dist, name)

    def met(ego):return met(ego._dist.sample(ego._samples), 0)

    def med(ego):
        sample = ego._dist.sample(ego._samples)
        logprob = ego._dist.log_prob(sample)
        return sample[mix(logprob)][0]

    def entropy(ego):
        sample = ego._dist.sample(ego._samples)
        logprob = ego.log_prob(sample)
        return -met(logprob, 0)





class DiscDist:
    def __init__(
        ego,
        logits,
        low=-20.0,
        high=20.0,
        transfwd=symlog,
        transbwd=symexp,
        device="cuda",
    ):
        ego.logits = logits
        ego.probs = mar(logits, -1)
        ego.buckets = lan(low, high, steps=255).to(device)
        ego.width = (ego.buckets[-1] - ego.buckets[0]) / 255
        ego.transfwd = transfwd
        ego.transbwd = transbwd

    def met(ego):return ego.transbwd(sum(ego.probs*ego.buckets,dim=-1,keepdim=True))
    med=met

    # Inside OneHotCategorical, log_prob is calculated using only max element in targets
    def lop(ego, x):
        x = ego.transfwd(x)
        # x(time, batch, 1)
        below = sum((ego.buckets <= x[..., None]).to(nmb.int32), dim=-1) - 1
        above = len(ego.buckets) - sum(
            (ego.buckets > x[..., None]).to(nmb.int32), dim=-1
        )
        # this is implemented using clip at the original repo as the gradients are not backpropagated for the out of limits.
        below = mad(below, 0, len(ego.buckets) - 1)
        above = mad(above, 0, len(ego.buckets) - 1)
        equal = below == above

        dist_to_below = nmb.where(equal, 1, abs(ego.buckets[below] - x))
        dist_to_above = nmb.where(equal, 1, abs(ego.buckets[above] - x))
        total = dist_to_below + dist_to_above
        weight_below = dist_to_above / total
        weight_above = dist_to_below / total
        target = (
            nmf.one_hot(below, num_classes=len(ego.buckets)) * weight_below[..., None]
            + nmf.one_hot(above, num_classes=len(ego.buckets)) * weight_above[..., None]
        )
        log_pred = ego.logits - logsumexp(ego.logits, -1, keepdim=True)
        target = target.squeeze(-2)

        return (target * log_pred).sum(-1)

    def log_prob_target(ego, target):
        log_pred = super().logits - logsumexp(super().logits, -1, keepdim=True)
        return (target * log_pred).sum(-1)


class MSEDist:
    def __init__(ego, mode, agg="sum"):
        ego._mode = mode
        ego._agg = agg

    def met(ego):return ego._mode
    med=met

    def lop(ego, value):
        assert ego._mode.shape == value.shape, (ego._mode.shape, value.shape)
        distance = (ego._mode - value) ** 2
        if ego._agg == "mean":
            loss = distance.mean(list(DIT(distance.shape))[2:])
        elif ego._agg == "sum":
            loss = distance.sum(list(DIT(distance.shape))[2:])
        else:
            raise NotImplementedError(ego._agg)
        return -loss


class SymlogDist:
    def __init__(ego, mode, dist="mse", agg="sum", tol=1e-8):
        ego._mode = mode
        ego._dist = dist
        ego._agg = agg
        ego._tol = tol
    def met(ego):return symexp(ego._mode)
    med=met
    def lop(ego, value):
        assert ego._mode.shape == value.shape
        if ego._dist == "mse":
            distance = (ego._mode - symlog(value)) ** 2.0
            distance = nmb.where(distance < ego._tol, 0, distance)
        elif ego._dist == "abs":
            distance = abs(ego._mode - symlog(value))
            distance = nmb.where(distance < ego._tol, 0, distance)
        else:
            raise NotImplementedError(ego._dist)
        if ego._agg == "mean":
            loss = distance.mean(list(DIT(distance.shape))[2:])
        elif ego._agg == "sum":
            loss = distance.sum(list(DIT(distance.shape))[2:])
        else:
            raise NotImplementedError(ego._agg)
        return -loss


class ContDist:
    def __init__(ego, dist=None, absmax=None):
        super().__init__()
        ego._dist = dist
        ego.mean = dist.mean
        ego.absmax = absmax

    def __getattr__(ego, name):
        return getattr(ego._dist, name)

    def entropy(ego):
        return ego._dist.entropy()

    def med(ego):
        out = ego._dist.mean
        if ego.absmax is not None:
            out *= (ego.absmax / mad(abs(out), min=ego.absmax)).detach()
        return out

    def sample(ego, sample_shape=()):
        out = ego._dist.rsample(sample_shape)
        if ego.absmax is not None:
            out *= (ego.absmax / mad(abs(out), min=ego.absmax)).detach()
        return out

    def lop(ego, x):
        return ego._dist.log_prob(x)


class Bernoulli:
    def __init__(ego, dist=None):
        super().__init__()
        ego._dist = dist
        ego.mean = dist.mean

    def __getattr__(ego, name):
        return getattr(ego._dist, name)

    def entropy(ego):
        return ego._dist.entropy()

    def med(ego):
        _mode = rod(ego._dist.mean)
        return _mode.detach() + ego._dist.mean - ego._dist.mean.detach()

    def sample(ego, sample_shape=()):
        return ego._dist.rsample(sample_shape)

    def lop(ego, x):
        _logits = ego._dist.base_dist.logits
        log_probs0 = -F.softplus(_logits)
        log_probs1 = -F.softplus(-_logits)

        return sum(log_probs0 * (1 - x) + log_probs1 * x, -1)

# Categorical 离散概率分布
class Dal_Ctg(dal.Categorical):
  def met(ego):return ego.probs.argmax(dim=-1, keepdim=True)
  def sap(ego):return ego.sample().unsqueeze(-1)
  def lop(ego,iap):return ego.log_prob(iap.squeeze(-1)).view(iap.size(0), -1).sum(-1).unsqueeze(-1)
  
# Normal v ExponentialFamily
class Dal_Nrm(dal.Normal):
  def met(ego):return ego.mean
  def tro(ego):return ego.entropy().sum(-1)
  def lop(ego,iap):return ego.log_prob(iap)#.sum(-1, keepdim=True)[HAPPO BUG]
  
# Normal v ExponentialFamily diagonal Gaussian
class DalMNrm(dal.MultivariateNormal):
  def met(ego):return ego.mean
  def tro(ego):return ego.entropy().sum(-1)
  def lop(ego,iap):return ego.log_prob(iap)
  
# Bernoulli
class Dal_Brn(dal.Bernoulli):
  def met(ego):return nmb.gt(ego.probs, 0.5).float()
  def tro(ego):return ego.entropy().sum(-1)
  def lop(ego,iap):return ego.log_prob(iap).view(iap.size(0), -1).sum(-1).unsqueeze(-1)
  
class OneHotDist(dal.one_hot_categorical.OneHotCategorical):
    def __init__(ego, logits=None, probs=None, unimix_ratio=0.0):
        if logits is not None and unimix_ratio > 0.0:
            probs = mar(logits, dim=-1)
            probs = probs * (1.0 - unimix_ratio) + unimix_ratio / probs.shape[-1]
            logits = log(probs)
            super().__init__(logits=logits, probs=None)
        else:
            super().__init__(logits=logits, probs=probs)

    def med(ego):
        _mode = nmf.one_hot(
            mix(super().logits, axis=-1), super().logits.shape[-1]
        )
        return _mode.detach() + super().logits - super().logits.detach()

    def sample(ego, sample_shape=(), seed=None):
        if seed is not None:
            raise ValueError("need to check")
        sample = super().sample(sample_shape)
        probs = super().probs
        while len(probs.shape) < len(sample.shape):
            probs = probs[None]
        sample += probs - probs.detach()
        return sample





class UnnormalizedHuber(dal.normal.Normal):
    def __init__(ego, loc, scale, threshold=1, **kwargs):
        super().__init__(loc, scale, **kwargs)
        ego._threshold = threshold

    def lop(ego, event):
        return -(
            rsq((event - ego.mean) ** 2 + ego._threshold**2)
            - ego._threshold
        )

    def med(ego):
        return ego.mean


class SafeTruncatedNormal(dal.normal.Normal):
    def __init__(ego, loc, scale, low, high, clip=1e-6, mult=1):
        super().__init__(loc, scale)
        ego._low = low
        ego._high = high
        ego._clip = clip
        ego._mult = mult

    def sample(ego, sample_shape):
        event = super().sample(sample_shape)
        if ego._clip:
            clipped = mad(event, ego._low + ego._clip, ego._high - ego._clip)
            event = event - event.detach() + clipped.detach()
        if ego._mult:
            event *= ego._mult
        return event
    

Independent = dal.independent.Independent
Uniform = dal.uniform.Uniform


class Idd(dal.Distribution):
    #arg_constraints: Dict[str, constraints.Constraint] = {}
    def __init__(
        ego, base_distribution, reinterpreted_batch_ndims, validate_args=False #None
    ):
        if reinterpreted_batch_ndims > len(base_distribution.batch_shape):
            raise ValueError(
                "Expected reinterpreted_batch_ndims <= len(base_distribution.batch_shape), "
                f"actual {reinterpreted_batch_ndims} vs {len(base_distribution.batch_shape)}"
            )
        shape = base_distribution.batch_shape + base_distribution.event_shape
        event_dim = reinterpreted_batch_ndims + len(base_distribution.event_shape)
        batch_shape = shape[: len(shape) - event_dim]
        event_shape = shape[len(shape) - event_dim :]
        ego.base_dist = base_distribution
        ego.reinterpreted_batch_ndims = reinterpreted_batch_ndims
        
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    def expand(ego, batch_shape, _instance=None):
        new = ego._get_checked_instance(Independent, _instance)
        batch_shape = torch.Size(batch_shape)
        new.base_dist = ego.base_dist.expand(
            batch_shape + ego.event_shape[: ego.reinterpreted_batch_ndims]
        )
        new.reinterpreted_batch_ndims = ego.reinterpreted_batch_ndims
        super(Independent, new).__init__(
            batch_shape, ego.event_shape, validate_args=False
        )
        new._validate_args = ego._validate_args
        return new

    @property
    def has_rsample(ego):
        return ego.base_dist.has_rsample

    @property
    def has_enumerate_support(ego):
        if ego.reinterpreted_batch_ndims > 0:
            return False
        return ego.base_dist.has_enumerate_support

    @dal.constraints.dependent_property
    def support(ego):
        result = ego.base_dist.support
        if ego.reinterpreted_batch_ndims:
            result = constraints.independent(result, ego.reinterpreted_batch_ndims)
        return result

    #@property
    def met(ego):return ego.base_dist.mean

    #@property
    def med(ego):return ego.base_dist.mode

    #@property
    def mev(ego):return ego.base_dist.variance

    #sample
    def sap(ego, sample_shape=nmb.Size()):return ego.base_dist.sample(sample_shape)
    #rsample
    def rap(ego, sample_shape=nmb.Size()):
        return ego.base_dist.rsample(sample_shape)

    #log_prob
    def lop(ego, value):return _sum_rightmost(ego.base_dist.log_prob(value), ego.reinterpreted_batch_ndims)
    #entropy
    def tro(ego):return _sum_rightmost(ego.base_dist.entropy(), ego.reinterpreted_batch_ndims)

    def enumerate_support(ego, expand=True):
        if ego.reinterpreted_batch_ndims > 0:
            raise NotImplementedError("Enumeration over cartesian product is not implemented")
        return ego.base_dist.enumerate_support(expand=expand)

    def __repr__(ego):return ego.__class__.__name__+f"({ego.base_dist}, {ego.reinterpreted_batch_ndims})"