"""This module implements various mathematical sets like samlpable-space.
样本空间（通常用 S 或 Ω 表示）是指在进行一次随机实验或观察一个随机现象时，可能出现的全部基本结果（或称样本点）构成的集合。每一个基本结果都是实验或现象的一个具体、明确、互斥的可能状态。样本空间是描述随机性质问题的基础框架，包含了实验或观察的全部可能性。
关键特征:
- 完整性：样本空间包含了实验或现象的所有可能结果，没有遗漏。这意味着在给定的条件下，实验或现象的实际结果必定属于样本空间的一部分。
- 互斥性：样本空间中的每一个基本结果都是彼此排斥的，即在一次实验中，不能同时出现两个或更多的基本结果。例如，抛掷一枚公平的硬币时，样本空间是 {正面, 反面}，一次抛掷只能出现“正面”或“反面”，不可能同时出现两者。
- 基本事件：样本空间中的元素（基本结果）也被称为基本事件。基本事件是最小的、不可再分的事件单位，它们不能再进一步分解为更简单的事件。例如，在抛掷骰子的实验中，每个点数（如“掷出1点”、“掷出2点”等）就是一个基本事件。
- 离散与连续：样本空间根据其元素的性质可以分为离散样本空间和连续样本空间。离散样本空间的元素是可以一一列举的，如掷骰子或抛硬币的样本空间。连续样本空间的元素则构成一个区间或区域，如测量某一物体长度时，样本空间可以是所有实数（或某一有限区间内的实数），这样的样本空间无法一一列举，通常借助概率密度函数来描述。
- 多维样本空间：对于涉及多个随机变量的复杂实验或现象，样本空间可能需要通过多个基本变量的联合结果来定义。这时，样本空间可能是多个单个样本空间（每个对应一个随机变量）的笛卡尔积，每个元素代表所有变量值的特定组合。例如，从一副不含鬼牌的标准扑克牌中抽取一张牌，样本空间可以是数字（A到K）和花色（黑桃、红桃、梅花、方块）的笛卡尔积，完整描述一张牌需要同时指明数字和花色。
- 子集与事件：样本空间的任意子集被称为一个事件。事件可以是单个基本事件，也可以是多个基本事件的并集、交集等组合。事件的概率是基于样本空间中各基本事件的概率计算得出的。例如，“掷骰子得到偶数点”是一个事件，它包含样本空间中的两个基本事件：“掷出2点”、“掷出4点”、“掷出6点”。
MDP 中的特殊空间:
- spc_act:     spc_act[ego.spc_act = num.Discrete(3)]
- spc_obs:spc_obs[ego.spc_obs = spaces.Box(0, 1, shape=(2,))]
- spc_stt:      state_space[ego.spc_stt = num.Dict()]
"""
from.__deps_ import*
from.spc_box import*#Box
from.spc_dic import*#Spc_Dict
from.spc_dsc import*#Discrete, MultiDiscrete
from.spc_grf import*#Graph, GraphInstance
from.spc_mbi import*#MultiBinary
from.spc_seq import*#Sequence
from.spc_txt import*#Text
from.spc_tup import*#Tuple,Spc_NyH
from.utils import flatdim, flatten, flatten_space, unflatten

__all__ = [
    "Spc",
    "Box",
    "Spc_NyH",
    "Discrete",
    "Text",
    "Graph",
    "GraphInstance",
    "MultiDiscrete",
    "MultiBinary",
    "Tuple",
    "Sequence",
    "Spc_Dic",
    "flatdim",
    "flatten_space",
    "flatten",
    "unflatten",
]